@@ -40,6 +40,107 @@
   :link '(url-link :tag "Denote homepage" "https://protesilaos.com/emacs/denote")
   :link '(url-link :tag "Denote Org homepage" "https://protesilaos.com/emacs/denote-org"))
 
+(defconst denote-org--common-extra-parameters
+  '((list :inline t
+         :not-regexp
+         (choice (const :tag "Do not exclude anything" nil)
+                 (string :tag "Regular expression to exclude")))
+   (list :inline t
+         :excluded-dirs-regexp
+         (choice (const :tag "Do not exclude directories" nil)
+                 (string :tag "Regular expression to exclude directories")))
+   (list :inline t
+         :sort-by-component
+         ;; TODO 2025-07-30: Check the other sort options we provide.
+         (choice (const :tag "Sort by title" title)
+                 (const :tag "Sort by signature" signature)
+                 (const :tag "Sort by keywords" keywords)))
+   (list :inline t
+         :reverse-sort
+         (choice (const :tag "No reverse sort" nil)
+                 (const :tag "Reverse sort" t)))
+   (list :inline t
+         :id-only
+         (choice (const :tag "Links with full description" nil)
+                 (const :tag "Links with just the identifier" t)))
+   (list :inline t
+         :include-date
+         (choice (const :tag "Do not show the date" nil)
+                 (const :tag "Show the date" t))))
+  "For use in the :type of `denote-org-extra-parameters-alist'.")
+
+(defcustom denote-org-extra-parameters-alist
+  '((denote-links :not-regexp nil :excluded-dirs-regexp nil :sort-by-component nil :reverse-sort nil :id-only nil :include-date nil)
+    (denote-missing-links :not-regexp nil :excluded-dirs-regexp nil :sort-by-component nil :reverse-sort nil :id-only nil :include-date nil)
+    (denote-backlinks :not-regexp nil :excluded-dirs-regexp nil :sort-by-component nil :reverse-sort nil :id-only nil :this-heading-only nil :include-date nil)
+    (denote-files :not-regexp nil :excluded-dirs-regexp nil :sort-by-component title :reverse-sort nil :no-front-matter nil :file-separator nil :add-links nil)
+    (denote-files-as-headings :not-regexp nil :excluded-dirs-regexp nil :sort-by-component title :reverse-sort nil :add-links nil))
+  "Extra parameters for each Denote Org dynamic block.
+Those are in addition to the mandatory parameters that will always be
+present in a dynamic block."
+  :type '(set
+          (list
+           (list :inline t
+                 (const denote-links)
+                 (set
+                  (list :inline t
+                        :not-regexp
+                        (choice (const :tag "Do not exclude anything" nil)
+                                (string :tag "Regular expression to exclude")))
+                  (list :inline t
+                        :excluded-dirs-regexp
+                        (choice (const :tag "Do not exclude directories" nil)
+                                (string :tag "Regular expression to exclude directories")))
+                  (list :inline t
+                        :sort-by-component
+                        ;; TODO 2025-07-30: Check the other sort options we provide.
+                        (choice (const :tag "Sort by title" title)
+                                (const :tag "Sort by signature" signature)
+                                (const :tag "Sort by keywords" keywords)))
+                  (list :inline t
+                        :reverse-sort
+                        (choice (const :tag "No reverse sort" nil)
+                                (const :tag "Reverse sort" t)))
+                  (list :inline t
+                        :id-only
+                        (choice (const :tag "Links with full description" nil)
+                                (const :tag "Links with just the identifier" t)))
+                  (list :inline t
+                        :include-date
+                        (choice (const :tag "Do not show the date" nil)
+                                (const :tag "Show the date" t)))))
+
+           (list :inline t
+                 (const denote-missing-links)
+                 (set
+                  (list :inline t
+                        :excluded-dirs-regexp
+                        (choice (const :tag "Do not exclude directories" nil)
+                                (string :tag "Regular expression to exclude directories")))
+                  (list :inline t
+                        :sort-by-component
+                        (choice (const :tag "Sort by title" title)
+                                (const :tag "Sort by title" signature)
+                                (const :tag "Sort by title" keywords)))
+                  (list :inline t
+                        :reverse-sort
+                        (choice (const :tag "No reverse sort" nil)
+                                (const :tag "Reverse sort" t)))
+                  (list :inline t
+                        :id-only
+                        (choice (const :tag "Links with full description" nil)
+                                (const :tag "Links with just the identifier" t)))
+                  (list :inline t
+                        :include-date
+                        (choice (const :tag "Do not show the date" nil)
+                                (const :tag "Show the date" t)))))))
+                             
+           ;; (const denote-missing-links)
+           ;; (const denote-backlinks)
+           ;; (const denote-files)
+           ;; (const denote-files-as-headings))
+  :group 'denote-org)
+
 ;;;; Link to file and heading
 
 (defun denote-org--get-outline (file)
@@ -384,9 +485,9 @@ point to a file with a Denote file name."
 ;;    (info "(org) Dynamic Blocks")
 ;;
 ;; The dynamic blocks defined herein are documented at length in the
-;; Denote manual.  See the following node and its subsections:
+;; Denote-Org manual.  See the following node and its subsections:
 ;;
-;;    (info "(denote) Use Org dynamic blocks")
+;;    (info "(denote-org) Use Org dynamic blocks")
 
 ;;;;; Common helper functions
 
@@ -407,26 +508,27 @@ Also see `denote-org-dblock--files-missing-only'."
    (t
     (denote-directory-files files-matching-regexp :omit-current nil exclude-regexp))))
 
-(defun denote-org-dblock--get-missing-links (regexp)
+(defun denote-org-dblock--get-missing-links (regexp exclude-regexp)
   "Return list of missing links to all notes matching REGEXP.
 Missing links are those for which REGEXP does not have a match in
-the current buffer."
-  (let ((found-files (denote-directory-files regexp :omit-current))
-        (linked-files (denote-link--expand-identifiers denote-org-link-in-context-regexp)))
-    (if-let* ((final-files (seq-difference found-files linked-files)))
-        final-files
-      (message "All links matching `%s' are present" regexp)
-      '())))
+the current buffer.
 
-(defun denote-org-dblock--files-missing-only (files-matching-regexp &optional sort-by-component reverse)
+EXCLUDE-REGEXP is filtered out of the matching files."
+  (when-let* ((all-files (denote-directory-files regexp :omit-current nil exclude-regexp))
+              (linked-files (denote-get-links nil all-files)))
+    (seq-difference all-files linked-files)))
+
+(defun denote-org-dblock--files-missing-only (files-matching-regexp &optional sort-by-component reverse exclude-regexp)
   "Return list of missing links to FILES-MATCHING-REGEXP.
 SORT-BY-COMPONENT and REVERSE have the same meaning as
 `denote-sort-files'.  If both are nil, do not try to perform any
 sorting.
 
+EXCLUDE-REGEXP is filtered out of the matching files.
+
 Also see `denote-org-dblock--files'."
   (denote-sort-files
-   (denote-org-dblock--get-missing-links files-matching-regexp)
+   (denote-org-dblock--get-missing-links files-matching-regexp exclude-regexp)
    sort-by-component
    reverse))
 
@@ -454,12 +556,6 @@ Also see `denote-org-dblock--files'."
                            :include-date nil))
   (org-update-dblock))
 
-;; NOTE 2024-03-30: This is how the autoload is done in org.el.
-;;;###autoload
-(eval-after-load 'org
-  '(progn
-     (org-dynamic-block-define "denote-links" 'denote-org-dblock-insert-links)))
-
 ;; TODO 2024-12-04: Maybe we can do this for anything that deals with
 ;; regular expressions that users provide?  I prefer not to do the
 ;; work if nobody wants it, though I am mentioning this here just in
@@ -484,21 +580,16 @@ Used by `org-dblock-update' with PARAMS provided by the dynamic block."
          (not-rx (denote-org--parse-rx (plist-get params :not-regexp)))
          (sort (plist-get params :sort-by-component))
          (reverse (plist-get params :reverse-sort))
-         (include-date (plist-get params :include-date))
          (block-name (plist-get params :block-name))
          (denote-excluded-directories-regexp (or (plist-get params :excluded-dirs-regexp)
                                                  denote-excluded-directories-regexp))
          (files (denote-org-dblock--files rx sort reverse not-rx)))
     (when block-name (insert "#+name: " block-name "\n"))
-    (denote-link--insert-links files 'org (plist-get params :id-only) :no-other-sorting include-date)
+    (denote-org--insert-links files (plist-get params :id-only) (plist-get params :include-date))
     (join-line))) ; remove trailing empty line
 
 ;;;;; Dynamic block to insert missing links
 
-;; TODO 2024-12-03: Do we need the :not-regexp here?  I think yes,
-;; though I prefer to have a user of this kind of dblock send me their
-;; feedback.
-
 (define-obsolete-function-alias
   'denote-org-extras-dblock-insert-missing-links
   'denote-org-dblock-insert-missing-links
@@ -513,6 +604,7 @@ Used by `org-dblock-update' with PARAMS provided by the dynamic block."
    org-mode)
   (org-create-dblock (list :name "denote-missing-links"
                            :regexp regexp
+                           :not-regexp nil
                            :excluded-dirs-regexp nil
                            :sort-by-component nil
                            :reverse-sort nil
@@ -520,26 +612,20 @@ Used by `org-dblock-update' with PARAMS provided by the dynamic block."
                            :include-date nil))
   (org-update-dblock))
 
-;; NOTE 2024-03-30: This is how the autoload is done in org.el.
-;;;###autoload
-(eval-after-load 'org
-  '(progn
-     (org-dynamic-block-define "denote-missing-links" 'denote-org-dblock-insert-missing-links)))
-
 ;;;###autoload
 (defun org-dblock-write:denote-missing-links (params)
   "Function to update `denote-links' Org Dynamic blocks.
 Used by `org-dblock-update' with PARAMS provided by the dynamic block."
   (let* ((rx (denote-org--parse-rx (plist-get params :regexp)))
+         (not-rx (denote-org--parse-rx (plist-get params :not-regexp)))
          (sort (plist-get params :sort-by-component))
          (reverse (plist-get params :reverse-sort))
-         (include-date (plist-get params :include-date))
          (block-name (plist-get params :block-name))
          (denote-excluded-directories-regexp (or (plist-get params :excluded-dirs-regexp)
                                                  denote-excluded-directories-regexp))
-         (files (denote-org-dblock--files-missing-only rx sort reverse)))
+         (files (denote-org-dblock--files-missing-only rx sort reverse not-rx)))
     (when block-name (insert "#+name: " block-name "\n"))
-    (denote-link--insert-links files 'org (plist-get params :id-only) :no-other-sorting include-date)
+    (denote-org--insert-links files (plist-get params :id-only) (plist-get params :include-date))
     (join-line))) ; remove trailing empty line
 
 ;;;;; Dynamic block to insert backlinks
@@ -578,11 +664,27 @@ Used by `org-dblock-update' with PARAMS provided by the dynamic block."
                            :include-date nil))
   (org-update-dblock))
 
-;; NOTE 2024-03-30: This is how the autoload is done in org.el.
-;;;###autoload
-(eval-after-load 'org
-  '(progn
-     (org-dynamic-block-define "denote-backlinks" 'denote-org-dblock-insert-backlinks)))
+(defun denote-org--insert-links (files &optional id-only include-date)
+  "Insert links to FILES.
+With optional ID-ONLY, insert links with the corresponding file's
+identifier as the description.  Otherwise, use the title of the file.
+
+With optional INCLUDE-DATE, append the date of the file in (YYYY-MM-DD)
+format."
+  (let ((denote-link-description-format
+         (if include-date
+             ;; TODO 2025-05-21: Consider making this more flexible.
+             ;; The style we have now is what we have had since this
+             ;; feature was first introduced.  But maybe users prefer
+             ;; to have a different format.
+             (lambda (file)
+               (let* ((file-type (denote-filetype-heuristics file))
+                      (title (denote-retrieve-title-or-filename file file-type))
+                      (identifier (denote-retrieve-filename-identifier file))
+                      (date (denote-id-to-date identifier)))
+                 (format "%s (%s)" title date)))
+           denote-link-description-format)))
+    (denote-link--insert-links files 'org id-only :no-other-sorting)))
 
 ;;;###autoload
 (defun org-dblock-write:denote-backlinks (params)
@@ -590,18 +692,32 @@ Used by `org-dblock-update' with PARAMS provided by the dynamic block."
 Used by `org-dblock-update' with PARAMS provided by the dynamic block."
   (when-let* ((files (if (plist-get params :this-heading-only)
                          (denote-org--get-backlinks-for-heading (denote-org--get-file-id-and-heading-id-or-context))
-                       (denote-link-return-backlinks))))
-    (let* ((sort (plist-get params :sort-by-component))
+                       (denote-get-backlinks))))
+    (let* ((not-rx (plist-get params :not-regexp))
+           (sort (plist-get params :sort-by-component))
            (reverse (plist-get params :reverse-sort))
-           (include-date (plist-get params :include-date))
            (denote-excluded-directories-regexp (or (plist-get params :excluded-dirs-regexp)
                                                    denote-excluded-directories-regexp))
-           (files (denote-org-dblock--maybe-sort-backlinks files sort reverse)))
-      (denote-link--insert-links files 'org (plist-get params :id-only) :no-other-sorting include-date)
+           (files (denote-org-dblock--maybe-sort-backlinks files sort reverse))
+           (output (if not-rx
+                       (seq-remove
+                        (lambda (file)
+                          (string-match-p not-rx file))
+                        files)
+                     files)))
+      (denote-org--insert-links output (plist-get params :id-only) (plist-get params :include-date))
       (join-line)))) ; remove trailing empty line
 
 ;;;;; Dynamic block to insert entire file contents
 
+(defun denote-org-escape-code-in-region (beg end)
+  "Like `org-escape-code-in-region' to escape only #+ by appending a zero width space to it.
+Operate on the region between positions BEG and END."
+  (save-excursion
+    (goto-char end)
+    (while (re-search-backward "^[ \t]*\\(#\\+\\)" beg t)
+      (save-excursion (replace-match "​\\1" nil nil nil 1)))))
+
 (defun denote-org-dblock--get-file-contents (file &optional no-front-matter add-links)
   "Insert the contents of FILE.
 With optional NO-FRONT-MATTER as non-nil, try to remove the front
@@ -633,7 +749,8 @@ argument."
              (1+ (re-search-forward "^$" nil :no-error 1)))
            beginning-of-contents))
         (when add-links
-          (indent-region beginning-of-contents (point-max) 2)))
+          (indent-region beginning-of-contents (point-max) 2))
+        (denote-org-escape-code-in-region beginning-of-contents (point-max)))
       (buffer-string))))
 
 (defvar denote-org-dblock-file-contents-separator
@@ -707,12 +824,6 @@ among `denote-sort-components'."
                            :add-links nil))
   (org-update-dblock))
 
-;; NOTE 2024-03-30: This is how the autoload is done in org.el.
-;;;###autoload
-(eval-after-load 'org
-  '(progn
-     (org-dynamic-block-define "denote-files" 'denote-org-dblock-insert-files)))
-
 ;;;###autoload
 (defun org-dblock-write:denote-files (params)
   "Function to update `denote-files' Org Dynamic blocks.
@@ -734,10 +845,11 @@ Used by `org-dblock-update' with PARAMS provided by the dynamic block."
 
 (defun denote-org-dblock--extract-regexp (regexp)
   "Extract REGEXP from the buffer and trim it of surrounding spaces."
-  (string-trim
-   (save-excursion
-     (re-search-forward regexp nil :no-error)
-     (buffer-substring-no-properties (match-end 0) (line-end-position)))))
+  (when-let* ((_ (save-excursion (re-search-forward regexp nil :no-error)))
+              (match-end (match-end 0))
+              (line-end (line-end-position))
+              (text (buffer-substring-no-properties match-end line-end)))
+    (string-trim text)))
 
 (defun denote-org-dblock--get-file-contents-as-heading (file add-links)
   "Insert the contents of Org FILE, formatting the #+title as a heading.
@@ -753,7 +865,10 @@ With optional ADD-LINKS, make the title link to the original file."
         (insert-file-contents file)
         (setq title (denote-org-dblock--extract-regexp (denote--title-key-regexp file-type)))
         (setq tags (denote-org-dblock--extract-regexp (denote--keywords-key-regexp file-type)))
-        (delete-region (1+ (re-search-forward "^$" nil :no-error 1)) beginning-of-contents)
+        (when-let* ((_ (re-search-forward "^$" nil :no-error 1))
+                    (match-end (+ (match-end 0) 1))
+                    (_ (>= (point-max) match-end)))
+          (delete-region match-end beginning-of-contents))
         (goto-char beginning-of-contents)
         (when (and title tags)
           (if add-links
@@ -761,7 +876,8 @@ With optional ADD-LINKS, make the title link to the original file."
             (insert (format "* %s %s\n\n" title tags)))
           (org-align-tags :all))
         (while (re-search-forward "^\\(*+?\\) " nil :no-error)
-          (replace-match (format "*%s " "\\1"))))
+          (replace-match (format "*%s " "\\1")))
+        (denote-org-escape-code-in-region beginning-of-contents (point-max)))
       (buffer-string))))
 
 (defun denote-org-dblock-add-files-as-headings (regexp &optional add-links sort-by-component reverse excluded-dirs-regexp exclude-regexp)
@@ -828,12 +944,6 @@ as its own heading."
                            :add-links nil))
   (org-update-dblock))
 
-;; NOTE 2024-03-30: This is how the autoload is done in org.el.
-;;;###autoload
-(eval-after-load 'org
-  '(progn
-     (org-dynamic-block-define "denote-files-as-headings" 'denote-org-dblock-insert-files-as-headings)))
-
 ;;;###autoload
 (defun org-dblock-write:denote-files-as-headings (params)
   "Function to update `denote-files' Org Dynamic blocks.
@@ -849,5 +959,99 @@ Used by `org-dblock-update' with PARAMS provided by the dynamic block."
     (when rx (denote-org-dblock-add-files-as-headings rx add-links sort reverse excluded-dirs not-rx)))
   (join-line)) ; remove trailing empty line
 
+;;;;; Dynamic block to insert hierarchic sequences
+
+;; NOTE 2025-08-12: I am putting the autoloads here because I am now
+;; how they behave inside the `with-eval-after-load'.
+
+;;;###autoload (autoload 'denote-org-dblock-insert-sequence "denote-org")
+;;;###autoload (autoload 'org-dblock-write:denote-sequence "denote-org")
+
+(with-eval-after-load 'denote-sequence
+  (defun denote-org-dblock-insert-sequence (file depth)
+    "Create Org dynamic block to list all chilren of FILE up to a relative DEPTH.
+DEPTH of the root FILE is 1. Using 2 lists children, 3 grandchildren, and so on."
+    (interactive
+     (list
+      (denote-sequence-file-prompt
+       (format "List descendants of:"
+               (propertize
+                (denote--rename-dired-file-or-current-file-or-prompt)
+                'face 'denote-faces-prompt-current-name)))
+      (read-number "Maximum relative depth from root node: " 2))
+     (org-mode))
+    (org-create-dblock (list :name "denote-sequence"
+                             :sequence (denote-retrieve-filename-signature file)
+                             :depth depth))
+    (org-update-dblock))
+
+  (defun denote-org-sequence--get-files-with-max-depth (max-depth &optional files)
+    "Return members of FILES with sequence depth less or equal than MAX-DEPTH.
+When no FILES are provided, use all files with a sequence signature."
+    (unless files
+      (setq files (denote-sequence-get-all-files)))
+    (let* ((hierarchy '())
+           (files (denote-sequence-sort-files files))
+           (root-sequence (denote-retrieve-filename-signature (car files)))
+           (root-depth (denote-sequence-depth root-sequence)))
+      (message (number-to-string root-depth))
+      (dolist (file files)
+        (let* ((sequence (denote-retrieve-filename-signature file))
+               (depth (denote-sequence-depth sequence)))
+          (when (<= depth (- (+ root-depth max-depth) 1))
+            (push file hierarchy))))
+      (nreverse hierarchy)))
+
+  (defun org-dblock-write:denote-sequence (params)
+    "Function to update `denote-sequence' Org Dynamic blocks.
+When sequence is an empty string, then use all Denote files with a sequence.
+
+Used by `org-dblock-update' with PARAMS provided by the dynamic block."
+    (let* ((block-name (plist-get params :block-name))
+           (sequence (plist-get params :sequence))
+           (depth (plist-get params :depth))
+           ;; This will not work for people with bespoke `denote-file-name-components-order'
+           (parent (denote-directory-files (concat sequence "-")))
+           (children (denote-sequence-get-relative sequence 'all-children))
+           (family (if children
+                       (append parent children)
+                     (denote-sequence-get-all-files)))
+           (files (denote-org-sequence--get-files-with-max-depth depth family)))
+      (when block-name (insert "#+name: " block-name "\n"))
+      (denote-org--insert-sequence files)
+      (join-line)))
+
+  (defun denote-org--insert-sequence (files)
+    "Insert indented list of links to sequence FILES."
+    (let* ((root-sequence (denote-retrieve-filename-signature (car files)))
+           (root-depth (denote-sequence-depth root-sequence))
+           (links '()))
+      (message "Inserting %s links" (length files))
+      (dolist (file files)
+        (let* ((sequence (denote-retrieve-filename-signature file))
+               (description (denote-get-link-description file))
+               (link-title (concat sequence ": " description))
+               (link (denote-format-link file link-title 'org nil))
+               (depth (- (denote-sequence-depth sequence) root-depth))
+               (indent (make-string (* depth 2) ?\s))
+               (link-as-list-item (format (concat indent denote-link--prepare-links-format) link)))
+          (push link-as-list-item links)))
+      (setq links (nreverse links))
+      (dolist (link links)
+        (insert link)))))
+
+;;; 
+
+;; NOTE 2024-03-30: This is how the autoload is done in org.el.
+;;;###autoload
+(eval-after-load 'org
+  '(progn
+     (org-dynamic-block-define "denote-links" 'denote-org-dblock-insert-links)
+     (org-dynamic-block-define "denote-missing-links" 'denote-org-dblock-insert-missing-links)
+     (org-dynamic-block-define "denote-backlinks" 'denote-org-dblock-insert-backlinks)
+     (org-dynamic-block-define "denote-files" 'denote-org-dblock-insert-files)
+     (org-dynamic-block-define "denote-files-as-headings" 'denote-org-dblock-insert-files-as-headings)
+     (org-dynamic-block-define "denote-sequence" 'denote-org-dblock-insert-sequence)))
+
 (provide 'denote-org)
 ;;; denote-org.el ends here
